@@ -24,10 +24,11 @@
  * ~~~~
  */
 
-#include <netlink-local.h>
+#include <netlink-private/netlink.h>
 #include <netlink/netlink.h>
 #include <netlink/utils.h>
 #include <linux/socket.h>
+#include <stdlib.h> /* exit() */
 
 /**
  * Global variable indicating the desired level of debugging output.
@@ -48,6 +49,7 @@
 int nl_debug = 0;
 
 /** @cond SKIP */
+#ifdef NL_DEBUG
 struct nl_dump_params nl_debug_dp = {
 	.dp_type = NL_DUMP_DETAILS,
 };
@@ -64,6 +66,7 @@ static void __init nl_debug_init(void)
 
 	nl_debug_dp.dp_fd = stderr;
 }
+#endif
 
 int __nl_read_num_str_file(const char *path, int (*cb)(long, const char *))
 {
@@ -83,24 +86,32 @@ int __nl_read_num_str_file(const char *path, int (*cb)(long, const char *))
 			continue;
 
 		num = strtol(buf, &end, 0);
-		if (end == buf)
+		if (end == buf) {
+			fclose(fd);
 			return -NLE_INVAL;
+		}
 
-		if (num == LONG_MIN || num == LONG_MAX)
+		if (num == LONG_MIN || num == LONG_MAX) {
+			fclose(fd);
 			return -NLE_RANGE;
+		}
 
 		while (*end == ' ' || *end == '\t')
 			end++;
 
 		goodlen = strcspn(end, "#\r\n\t ");
-		if (goodlen == 0)
+		if (goodlen == 0) {
+			fclose(fd);
 			return -NLE_INVAL;
+		}
 
 		end[goodlen] = '\0';
 
 		err = cb(num, end);
-		if (err < 0)
+		if (err < 0) {
+			fclose(fd);
 			return err;
+		}
 	}
 
 	fclose(fd);
@@ -121,6 +132,7 @@ int __nl_read_num_str_file(const char *path, int (*cb)(long, const char *))
  *
  * Cancels down a byte counter until it reaches a reasonable
  * unit. The chosen unit is assigned to \a unit.
+ * This function assume 1024 bytes in one kilobyte
  * 
  * @return The cancelled down byte counter in the new unit.
  */
@@ -149,30 +161,36 @@ double nl_cancel_down_bytes(unsigned long long l, char **unit)
  * @arg	l		bit counter
  * @arg unit		destination unit pointer
  *
- * Cancels downa bit counter until it reaches a reasonable
+ * Cancels down bit counter until it reaches a reasonable
  * unit. The chosen unit is assigned to \a unit.
+ * This function assume 1000 bits in one kilobit
  *
  * @return The cancelled down bit counter in the new unit.
  */
 double nl_cancel_down_bits(unsigned long long l, char **unit)
 {
-	if (l >= 1099511627776ULL) {
+	if (l >= 1000000000000ULL) {
 		*unit = "Tbit";
-		return ((double) l) / 1099511627776ULL;
-	} else if (l >= 1073741824) {
-		*unit = "Gbit";
-		return ((double) l) / 1073741824;
-	} else if (l >= 1048576) {
-		*unit = "Mbit";
-		return ((double) l) / 1048576;
-	} else if (l >= 1024) {
-		*unit = "Kbit";
-		return ((double) l) / 1024;
-	} else {
-		*unit = "bit";
-		return (double) l;
+		return ((double) l) / 1000000000000ULL;
 	}
-		
+
+	if (l >= 1000000000) {
+		*unit = "Gbit";
+		return ((double) l) / 1000000000;
+	}
+
+	if (l >= 1000000) {
+		*unit = "Mbit";
+		return ((double) l) / 1000000;
+	}
+
+	if (l >= 1000) {
+		*unit = "Kbit";
+		return ((double) l) / 1000;
+	}
+
+	*unit = "bit";
+	return (double) l;
 }
 
 int nl_rate2str(unsigned long long rate, int type, char *buf, size_t len)
@@ -238,6 +256,9 @@ double nl_cancel_down_us(uint32_t l, char **unit)
  *  - b,kb/k,m/mb,gb/g for bytes
  *  - bit,kbit/mbit/gbit
  *
+ * This function assume 1000 bits in one kilobit and
+ * 1024 bytes in one kilobyte
+ *
  * @return The number of bytes or -1 if the string is unparseable
  */
 long nl_size2int(const char *str)
@@ -253,13 +274,13 @@ long nl_size2int(const char *str)
 		else if (!strcasecmp(p, "gb") || !strcasecmp(p, "g"))
 			l *= 1024*1024*1024;
 		else if (!strcasecmp(p, "gbit"))
-			l *= 1024*1024*1024/8;
+			l *= 1000000000L/8;
 		else if (!strcasecmp(p, "mb") || !strcasecmp(p, "m"))
 			l *= 1024*1024;
 		else if (!strcasecmp(p, "mbit"))
-			l *= 1024*1024/8;
+			l *= 1000000/8;
 		else if (!strcasecmp(p, "kbit"))
-			l *= 1024/8;
+			l *= 1000/8;
 		else if (!strcasecmp(p, "bit"))
 			l /= 8;
 		else if (strcasecmp(p, "b") != 0)
@@ -307,6 +328,11 @@ char *nl_size2str(const size_t size, char *buf, const size_t len)
 {
 	size_t i;
 
+	if (size == 0) {
+		snprintf(buf, len, "0B");
+		return buf;
+	}
+
 	for (i = 0; i < ARRAY_SIZE(size_units); i++) {
 		if (size >= size_units[i].limit) {
 			snprintf(buf, len, "%.2g%s",
@@ -411,10 +437,15 @@ static void __init get_psched_settings(void)
 			strncpy(name, "/proc/net/psched", sizeof(name) - 1);
 		
 		if ((fd = fopen(name, "r"))) {
-			uint32_t ns_per_usec, ns_per_tick, nom, denom;
+			unsigned int ns_per_usec, ns_per_tick, nom, denom;
 
-			fscanf(fd, "%08x %08x %08x %08x",
-			       &ns_per_usec, &ns_per_tick, &nom, &denom);
+			if (fscanf(fd, "%08x %08x %08x %08x",
+			       &ns_per_usec, &ns_per_tick, &nom, &denom) != 4) {
+                            NL_DBG(1, "Fatal error: can not read psched settings from \"%s\". " \
+                                    "Try to set TICKS_PER_USEC, PROC_NET_PSCHED or PROC_ROOT " \
+                                    "environment variables\n", name);
+                            exit(1);
+                        }
 
 			ticks_per_usec = (double) ns_per_usec / 
 					 (double) ns_per_tick;
@@ -520,6 +551,11 @@ char * nl_msec2str(uint64_t msec, char *buf, size_t len)
 	static const char *units[5] = {"d", "h", "m", "s", "msec"};
 	char * const buf_orig = buf;
 
+	if (msec == 0) {
+		snprintf(buf, len, "0msec");
+		return buf_orig;
+	}
+
 #define _SPLIT(idx, unit) if ((split[idx] = msec / unit)) msec %= unit
 	_SPLIT(0, 86400000);	/* days */
 	_SPLIT(1, 3600000);	/* hours */
@@ -549,24 +585,26 @@ char * nl_msec2str(uint64_t msec, char *buf, size_t len)
  */
 
 static const struct trans_tbl nlfamilies[] = {
-	__ADD(NETLINK_ROUTE,route)
-	__ADD(NETLINK_USERSOCK,usersock)
-	__ADD(NETLINK_FIREWALL,firewall)
-	__ADD(NETLINK_INET_DIAG,inetdiag)
-	__ADD(NETLINK_NFLOG,nflog)
-	__ADD(NETLINK_XFRM,xfrm)
-	__ADD(NETLINK_SELINUX,selinux)
-	__ADD(NETLINK_ISCSI,iscsi)
-	__ADD(NETLINK_AUDIT,audit)
-	__ADD(NETLINK_FIB_LOOKUP,fib_lookup)
-	__ADD(NETLINK_CONNECTOR,connector)
-	__ADD(NETLINK_NETFILTER,netfilter)
-	__ADD(NETLINK_IP6_FW,ip6_fw)
-	__ADD(NETLINK_DNRTMSG,dnrtmsg)
-	__ADD(NETLINK_KOBJECT_UEVENT,kobject_uevent)
-	__ADD(NETLINK_GENERIC,generic)
-	__ADD(NETLINK_SCSITRANSPORT,scsitransport)
-	__ADD(NETLINK_ECRYPTFS,ecryptfs)
+	__ADD(NETLINK_ROUTE,route),
+	__ADD(NETLINK_USERSOCK,usersock),
+	__ADD(NETLINK_FIREWALL,firewall),
+	__ADD(NETLINK_INET_DIAG,inetdiag),
+	__ADD(NETLINK_NFLOG,nflog),
+	__ADD(NETLINK_XFRM,xfrm),
+	__ADD(NETLINK_SELINUX,selinux),
+	__ADD(NETLINK_ISCSI,iscsi),
+	__ADD(NETLINK_AUDIT,audit),
+	__ADD(NETLINK_FIB_LOOKUP,fib_lookup),
+	__ADD(NETLINK_CONNECTOR,connector),
+	__ADD(NETLINK_NETFILTER,netfilter),
+	__ADD(NETLINK_IP6_FW,ip6_fw),
+	__ADD(NETLINK_DNRTMSG,dnrtmsg),
+	__ADD(NETLINK_KOBJECT_UEVENT,kobject_uevent),
+	__ADD(NETLINK_GENERIC,generic),
+	__ADD(NETLINK_SCSITRANSPORT,scsitransport),
+	__ADD(NETLINK_ECRYPTFS,ecryptfs),
+	__ADD(NETLINK_RDMA,rdma),
+	__ADD(NETLINK_CRYPTO,crypto),
 };
 
 char * nl_nlfamily2str(int family, char *buf, size_t size)
@@ -591,84 +629,83 @@ int nl_str2nlfamily(const char *name)
 
 static const struct trans_tbl llprotos[] = {
 	{0, "generic"},
-	__ADD(ARPHRD_ETHER,ether)
-	__ADD(ARPHRD_EETHER,eether)
-	__ADD(ARPHRD_AX25,ax25)
-	__ADD(ARPHRD_PRONET,pronet)
-	__ADD(ARPHRD_CHAOS,chaos)
-	__ADD(ARPHRD_IEEE802,ieee802)
-	__ADD(ARPHRD_ARCNET,arcnet)
-	__ADD(ARPHRD_APPLETLK,atalk)
-	__ADD(ARPHRD_DLCI,dlci)
-	__ADD(ARPHRD_ATM,atm)
-	__ADD(ARPHRD_METRICOM,metricom)
-	__ADD(ARPHRD_IEEE1394,ieee1394)
-#ifdef ARPHRD_EUI64
-	__ADD(ARPHRD_EUI64,eui64)
-#endif
-	__ADD(ARPHRD_INFINIBAND,infiniband)
-	__ADD(ARPHRD_SLIP,slip)
-	__ADD(ARPHRD_CSLIP,cslip)
-	__ADD(ARPHRD_SLIP6,slip6)
-	__ADD(ARPHRD_CSLIP6,cslip6)
-	__ADD(ARPHRD_RSRVD,rsrvd)
-	__ADD(ARPHRD_ADAPT,adapt)
-	__ADD(ARPHRD_ROSE,rose)
-	__ADD(ARPHRD_X25,x25)
-#ifdef ARPHRD_HWX25
-	__ADD(ARPHRD_HWX25,hwx25)
-#endif
-	__ADD(ARPHRD_CAN,can)
-	__ADD(ARPHRD_PPP,ppp)
-	__ADD(ARPHRD_HDLC,hdlc)
-	__ADD(ARPHRD_LAPB,lapb)
-	__ADD(ARPHRD_DDCMP,ddcmp)
-	__ADD(ARPHRD_RAWHDLC,rawhdlc)
-	__ADD(ARPHRD_TUNNEL,ipip)
-	__ADD(ARPHRD_TUNNEL6,tunnel6)
-	__ADD(ARPHRD_FRAD,frad)
-	__ADD(ARPHRD_SKIP,skip)
-	__ADD(ARPHRD_LOOPBACK,loopback)
-	__ADD(ARPHRD_LOCALTLK,localtlk)
-	__ADD(ARPHRD_FDDI,fddi)
-	__ADD(ARPHRD_BIF,bif)
-	__ADD(ARPHRD_SIT,sit)
-	__ADD(ARPHRD_IPDDP,ip/ddp)
-	__ADD(ARPHRD_IPGRE,gre)
-	__ADD(ARPHRD_PIMREG,pimreg)
-	__ADD(ARPHRD_HIPPI,hippi)
-	__ADD(ARPHRD_ASH,ash)
-	__ADD(ARPHRD_ECONET,econet)
-	__ADD(ARPHRD_IRDA,irda)
-	__ADD(ARPHRD_FCPP,fcpp)
-	__ADD(ARPHRD_FCAL,fcal)
-	__ADD(ARPHRD_FCPL,fcpl)
-	__ADD(ARPHRD_FCFABRIC,fcfb_0)
-	__ADD(ARPHRD_FCFABRIC+1,fcfb_1)
-	__ADD(ARPHRD_FCFABRIC+2,fcfb_2)
-	__ADD(ARPHRD_FCFABRIC+3,fcfb_3)
-	__ADD(ARPHRD_FCFABRIC+4,fcfb_4)
-	__ADD(ARPHRD_FCFABRIC+5,fcfb_5)
-	__ADD(ARPHRD_FCFABRIC+6,fcfb_6)
-	__ADD(ARPHRD_FCFABRIC+7,fcfb_7)
-	__ADD(ARPHRD_FCFABRIC+8,fcfb_8)
-	__ADD(ARPHRD_FCFABRIC+9,fcfb_9)
-	__ADD(ARPHRD_FCFABRIC+10,fcfb_10)
-	__ADD(ARPHRD_FCFABRIC+11,fcfb_11)
-	__ADD(ARPHRD_FCFABRIC+12,fcfb_12)
-	__ADD(ARPHRD_IEEE802_TR,tr)
-	__ADD(ARPHRD_IEEE80211,ieee802.11)
-	__ADD(ARPHRD_PHONET,phonet)
-	__ADD(ARPHRD_CAIF, caif)
-#ifdef ARPHRD_IEEE80211_PRISM
-	__ADD(ARPHRD_IEEE80211_PRISM, ieee802.11_prism)
-#endif
-#ifdef ARPHRD_VOID
-	__ADD(ARPHRD_VOID,void)
-#endif
-#ifdef ARPHRD_NONE
-	__ADD(ARPHRD_NONE,nohdr)
-#endif
+	__ADD(ARPHRD_NETROM,netrom),
+	__ADD(ARPHRD_ETHER,ether),
+	__ADD(ARPHRD_EETHER,eether),
+	__ADD(ARPHRD_AX25,ax25),
+	__ADD(ARPHRD_PRONET,pronet),
+	__ADD(ARPHRD_CHAOS,chaos),
+	__ADD(ARPHRD_IEEE802,ieee802),
+	__ADD(ARPHRD_ARCNET,arcnet),
+	__ADD(ARPHRD_APPLETLK,atalk),
+	__ADD(ARPHRD_DLCI,dlci),
+	__ADD(ARPHRD_ATM,atm),
+	__ADD(ARPHRD_METRICOM,metricom),
+	__ADD(ARPHRD_IEEE1394,ieee1394),
+	__ADD(ARPHRD_EUI64,eui64),
+	__ADD(ARPHRD_INFINIBAND,infiniband),
+	__ADD(ARPHRD_SLIP,slip),
+	__ADD(ARPHRD_CSLIP,cslip),
+	__ADD(ARPHRD_SLIP6,slip6),
+	__ADD(ARPHRD_CSLIP6,cslip6),
+	__ADD(ARPHRD_RSRVD,rsrvd),
+	__ADD(ARPHRD_ADAPT,adapt),
+	__ADD(ARPHRD_ROSE,rose),
+	__ADD(ARPHRD_X25,x25),
+	__ADD(ARPHRD_HWX25,hwx25),
+	__ADD(ARPHRD_CAN,can),
+	__ADD(ARPHRD_PPP,ppp),
+	__ADD(ARPHRD_CISCO,cisco),
+	__ADD(ARPHRD_HDLC,hdlc),
+	__ADD(ARPHRD_LAPB,lapb),
+	__ADD(ARPHRD_DDCMP,ddcmp),
+	__ADD(ARPHRD_RAWHDLC,rawhdlc),
+	__ADD(ARPHRD_TUNNEL,ipip),
+	__ADD(ARPHRD_TUNNEL6,tunnel6),
+	__ADD(ARPHRD_FRAD,frad),
+	__ADD(ARPHRD_SKIP,skip),
+	__ADD(ARPHRD_LOOPBACK,loopback),
+	__ADD(ARPHRD_LOCALTLK,localtlk),
+	__ADD(ARPHRD_FDDI,fddi),
+	__ADD(ARPHRD_BIF,bif),
+	__ADD(ARPHRD_SIT,sit),
+	__ADD(ARPHRD_IPDDP,ip/ddp),
+	__ADD(ARPHRD_IPGRE,gre),
+	__ADD(ARPHRD_PIMREG,pimreg),
+	__ADD(ARPHRD_HIPPI,hippi),
+	__ADD(ARPHRD_ASH,ash),
+	__ADD(ARPHRD_ECONET,econet),
+	__ADD(ARPHRD_IRDA,irda),
+	__ADD(ARPHRD_FCPP,fcpp),
+	__ADD(ARPHRD_FCAL,fcal),
+	__ADD(ARPHRD_FCPL,fcpl),
+	__ADD(ARPHRD_FCFABRIC,fcfb_0),
+	__ADD(ARPHRD_FCFABRIC+1,fcfb_1),
+	__ADD(ARPHRD_FCFABRIC+2,fcfb_2),
+	__ADD(ARPHRD_FCFABRIC+3,fcfb_3),
+	__ADD(ARPHRD_FCFABRIC+4,fcfb_4),
+	__ADD(ARPHRD_FCFABRIC+5,fcfb_5),
+	__ADD(ARPHRD_FCFABRIC+6,fcfb_6),
+	__ADD(ARPHRD_FCFABRIC+7,fcfb_7),
+	__ADD(ARPHRD_FCFABRIC+8,fcfb_8),
+	__ADD(ARPHRD_FCFABRIC+9,fcfb_9),
+	__ADD(ARPHRD_FCFABRIC+10,fcfb_10),
+	__ADD(ARPHRD_FCFABRIC+11,fcfb_11),
+	__ADD(ARPHRD_FCFABRIC+12,fcfb_12),
+	__ADD(ARPHRD_IEEE802_TR,tr),
+	__ADD(ARPHRD_IEEE80211,ieee802.11),
+	__ADD(ARPHRD_IEEE80211_PRISM,ieee802.11_prism),
+	__ADD(ARPHRD_IEEE80211_RADIOTAP,ieee802.11_radiotap),
+	__ADD(ARPHRD_IEEE802154,ieee802.15.4),
+	__ADD(ARPHRD_IEEE802154_MONITOR,ieee802.15.4_monitor),
+	__ADD(ARPHRD_PHONET,phonet),
+	__ADD(ARPHRD_PHONET_PIPE,phonet_pipe),
+	__ADD(ARPHRD_CAIF,caif),
+	__ADD(ARPHRD_IP6GRE,ip6gre),
+	__ADD(ARPHRD_NETLINK,netlink),
+	__ADD(ARPHRD_6LOWPAN,6lowpan),
+	__ADD(ARPHRD_VOID,void),
+	__ADD(ARPHRD_NONE,nohdr),
 };
 
 char * nl_llproto2str(int llproto, char *buf, size_t len)
@@ -690,75 +727,75 @@ int nl_str2llproto(const char *name)
  */
 
 static const struct trans_tbl ether_protos[] = {
-	__ADD(ETH_P_LOOP,loop)
-	__ADD(ETH_P_PUP,pup)
-	__ADD(ETH_P_PUPAT,pupat)
-	__ADD(ETH_P_IP,ip)
-	__ADD(ETH_P_X25,x25)
-	__ADD(ETH_P_ARP,arp)
-	__ADD(ETH_P_BPQ,bpq)
-	__ADD(ETH_P_IEEEPUP,ieeepup)
-	__ADD(ETH_P_IEEEPUPAT,ieeepupat)
-	__ADD(ETH_P_DEC,dec)
-	__ADD(ETH_P_DNA_DL,dna_dl)
-	__ADD(ETH_P_DNA_RC,dna_rc)
-	__ADD(ETH_P_DNA_RT,dna_rt)
-	__ADD(ETH_P_LAT,lat)
-	__ADD(ETH_P_DIAG,diag)
-	__ADD(ETH_P_CUST,cust)
-	__ADD(ETH_P_SCA,sca)
-	__ADD(ETH_P_TEB,teb)
-	__ADD(ETH_P_RARP,rarp)
-	__ADD(ETH_P_ATALK,atalk)
-	__ADD(ETH_P_AARP,aarp)
+	__ADD(ETH_P_LOOP,loop),
+	__ADD(ETH_P_PUP,pup),
+	__ADD(ETH_P_PUPAT,pupat),
+	__ADD(ETH_P_IP,ip),
+	__ADD(ETH_P_X25,x25),
+	__ADD(ETH_P_ARP,arp),
+	__ADD(ETH_P_BPQ,bpq),
+	__ADD(ETH_P_IEEEPUP,ieeepup),
+	__ADD(ETH_P_IEEEPUPAT,ieeepupat),
+	__ADD(ETH_P_DEC,dec),
+	__ADD(ETH_P_DNA_DL,dna_dl),
+	__ADD(ETH_P_DNA_RC,dna_rc),
+	__ADD(ETH_P_DNA_RT,dna_rt),
+	__ADD(ETH_P_LAT,lat),
+	__ADD(ETH_P_DIAG,diag),
+	__ADD(ETH_P_CUST,cust),
+	__ADD(ETH_P_SCA,sca),
+	__ADD(ETH_P_TEB,teb),
+	__ADD(ETH_P_RARP,rarp),
+	__ADD(ETH_P_ATALK,atalk),
+	__ADD(ETH_P_AARP,aarp),
 #ifdef ETH_P_8021Q
-	__ADD(ETH_P_8021Q,802.1q)
+	__ADD(ETH_P_8021Q,802.1q),
 #endif
-	__ADD(ETH_P_IPX,ipx)
-	__ADD(ETH_P_IPV6,ipv6)
-	__ADD(ETH_P_PAUSE,pause)
-	__ADD(ETH_P_SLOW,slow)
+	__ADD(ETH_P_IPX,ipx),
+	__ADD(ETH_P_IPV6,ipv6),
+	__ADD(ETH_P_PAUSE,pause),
+	__ADD(ETH_P_SLOW,slow),
 #ifdef ETH_P_WCCP
-	__ADD(ETH_P_WCCP,wccp)
+	__ADD(ETH_P_WCCP,wccp),
 #endif
-	__ADD(ETH_P_PPP_DISC,ppp_disc)
-	__ADD(ETH_P_PPP_SES,ppp_ses)
-	__ADD(ETH_P_MPLS_UC,mpls_uc)
-	__ADD(ETH_P_MPLS_MC,mpls_mc)
-	__ADD(ETH_P_ATMMPOA,atmmpoa)
-	__ADD(ETH_P_LINK_CTL,link_ctl)
-	__ADD(ETH_P_ATMFATE,atmfate)
-	__ADD(ETH_P_PAE,pae)
-	__ADD(ETH_P_AOE,aoe)
-	__ADD(ETH_P_TIPC,tipc)
-	__ADD(ETH_P_1588,ieee1588)
-	__ADD(ETH_P_FCOE,fcoe)
-	__ADD(ETH_P_FIP,fip)
-	__ADD(ETH_P_EDSA,edsa)
-	__ADD(ETH_P_EDP2,edp2)
-	__ADD(ETH_P_802_3,802.3)
-	__ADD(ETH_P_AX25,ax25)
-	__ADD(ETH_P_ALL,all)
-	__ADD(ETH_P_802_2,802.2)
-	__ADD(ETH_P_SNAP,snap)
-	__ADD(ETH_P_DDCMP,ddcmp)
-	__ADD(ETH_P_WAN_PPP,wan_ppp)
-	__ADD(ETH_P_PPP_MP,ppp_mp)
-	__ADD(ETH_P_LOCALTALK,localtalk)
-	__ADD(ETH_P_CAN,can)
-	__ADD(ETH_P_PPPTALK,ppptalk)
-	__ADD(ETH_P_TR_802_2,tr_802.2)
-	__ADD(ETH_P_MOBITEX,mobitex)
-	__ADD(ETH_P_CONTROL,control)
-	__ADD(ETH_P_IRDA,irda)
-	__ADD(ETH_P_ECONET,econet)
-	__ADD(ETH_P_HDLC,hdlc)
-	__ADD(ETH_P_ARCNET,arcnet)
-	__ADD(ETH_P_DSA,dsa)
-	__ADD(ETH_P_TRAILER,trailer)
-	__ADD(ETH_P_PHONET,phonet)
-	__ADD(ETH_P_IEEE802154,ieee802154)
-	__ADD(ETH_P_CAIF,caif)
+	__ADD(ETH_P_PPP_DISC,ppp_disc),
+	__ADD(ETH_P_PPP_SES,ppp_ses),
+	__ADD(ETH_P_MPLS_UC,mpls_uc),
+	__ADD(ETH_P_MPLS_MC,mpls_mc),
+	__ADD(ETH_P_ATMMPOA,atmmpoa),
+	__ADD(ETH_P_LINK_CTL,link_ctl),
+	__ADD(ETH_P_ATMFATE,atmfate),
+	__ADD(ETH_P_PAE,pae),
+	__ADD(ETH_P_AOE,aoe),
+	__ADD(ETH_P_TIPC,tipc),
+	__ADD(ETH_P_1588,ieee1588),
+	__ADD(ETH_P_FCOE,fcoe),
+	__ADD(ETH_P_FIP,fip),
+	__ADD(ETH_P_EDSA,edsa),
+	__ADD(ETH_P_EDP2,edp2),
+	__ADD(ETH_P_802_3,802.3),
+	__ADD(ETH_P_AX25,ax25),
+	__ADD(ETH_P_ALL,all),
+	__ADD(ETH_P_802_2,802.2),
+	__ADD(ETH_P_SNAP,snap),
+	__ADD(ETH_P_DDCMP,ddcmp),
+	__ADD(ETH_P_WAN_PPP,wan_ppp),
+	__ADD(ETH_P_PPP_MP,ppp_mp),
+	__ADD(ETH_P_LOCALTALK,localtalk),
+	__ADD(ETH_P_CAN,can),
+	__ADD(ETH_P_PPPTALK,ppptalk),
+	__ADD(ETH_P_TR_802_2,tr_802.2),
+	__ADD(ETH_P_MOBITEX,mobitex),
+	__ADD(ETH_P_CONTROL,control),
+	__ADD(ETH_P_IRDA,irda),
+	__ADD(ETH_P_ECONET,econet),
+	__ADD(ETH_P_HDLC,hdlc),
+	__ADD(ETH_P_ARCNET,arcnet),
+	__ADD(ETH_P_DSA,dsa),
+	__ADD(ETH_P_TRAILER,trailer),
+	__ADD(ETH_P_PHONET,phonet),
+	__ADD(ETH_P_IEEE802154,ieee802154),
+	__ADD(ETH_P_CAIF,caif),
 };
 
 char *nl_ether_proto2str(int eproto, char *buf, size_t len)
@@ -965,7 +1002,7 @@ char *__flags2str(int flags, char *buf, size_t len,
 	int tmp = flags;
 
 	memset(buf, 0, len);
-	
+
 	for (i = 0; i < tbl_len; i++) {
 		if (tbl[i].i & tmp) {
 			tmp &= ~tbl[i].i;
@@ -1073,6 +1110,81 @@ void dump_from_ops(struct nl_object *obj, struct nl_dump_params *params)
 		obj->ce_ops->oo_dump[type](obj, params);
 }
 
+/**
+ * Check for library capabilities
+ *
+ * @arg	capability	capability identifier
+ *
+ * Check whether the loaded libnl library supports a certain capability.
+ * This is useful so that applications can workaround known issues of
+ * libnl that are fixed in newer library versions, without
+ * having a hard dependency on the new version. It is also useful, for
+ * capabilities that cannot easily be detected using autoconf tests.
+ * The capabilities are integer constants with name NL_CAPABILITY_*.
+ *
+ * As this function is intended to detect capabilities at runtime,
+ * you might not want to depend during compile time on the NL_CAPABILITY_*
+ * names. Instead you can use their numeric values which are guaranteed not to
+ * change meaning.
+ *
+ * @return non zero if libnl supports a certain capability, 0 otherwise.
+ **/
+int nl_has_capability (int capability)
+{
+	static const uint8_t caps[ ( NL_CAPABILITY_MAX + 7 ) / 8  ] = {
+#define _NL_ASSERT(expr) ( 0 * sizeof(struct { unsigned int x: ( (!!(expr)) ? 1 : -1 ); }) )
+#define _NL_SETV(i, r, v) \
+		( _NL_ASSERT( (v) == 0 || (i) * 8 + (r) == (v) - 1 ) + \
+		  ( (v) == 0 ? 0 : (1 << (r)) ) )
+#define _NL_SET(i, v0, v1, v2, v3, v4, v5, v6, v7) \
+		[(i)] = ( \
+			_NL_SETV((i), 0, (v0)) | _NL_SETV((i), 4, (v4)) | \
+			_NL_SETV((i), 1, (v1)) | _NL_SETV((i), 5, (v5)) | \
+			_NL_SETV((i), 2, (v2)) | _NL_SETV((i), 6, (v6)) | \
+			_NL_SETV((i), 3, (v3)) | _NL_SETV((i), 7, (v7)) )
+		_NL_SET(0,
+			NL_CAPABILITY_ROUTE_BUILD_MSG_SET_SCOPE,
+			NL_CAPABILITY_ROUTE_LINK_VETH_GET_PEER_OWN_REFERENCE,
+			NL_CAPABILITY_ROUTE_LINK_CLS_ADD_ACT_OWN_REFERENCE,
+			NL_CAPABILITY_NL_CONNECT_RETRY_GENERATE_PORT_ON_ADDRINUSE,
+			NL_CAPABILITY_ROUTE_LINK_GET_KERNEL_FAIL_OPNOTSUPP,
+			NL_CAPABILITY_ROUTE_ADDR_COMPARE_CACHEINFO,
+			NL_CAPABILITY_VERSION_3_2_26,
+			NL_CAPABILITY_NL_RECV_FAIL_TRUNC_NO_PEEK),
+		_NL_SET(1,
+			NL_CAPABILITY_LINK_BUILD_CHANGE_REQUEST_SET_CHANGE,
+			NL_CAPABILITY_RTNL_NEIGH_GET_FILTER_AF_UNSPEC_FIX,
+			NL_CAPABILITY_VERSION_3_2_27,
+			0,
+			0,
+			0,
+			0,
+			0),
+		/* IMPORTANT: these capability numbers are intended to be universal and stable
+		 * for libnl3. Don't allocate new numbers on your own that differ from upstream
+		 * libnl3.
+		 *
+		 * Instead register a capability number upstream too. We will take patches
+		 * for that. We especially take patches to register a capability number that is
+		 * only implemented in your fork of libnl3.
+		 *
+		 * If you really don't want that, use capabilities in the range 0x7000 to 0x7FFF.
+		 * (NL_CAPABILITY_IS_USER_RESERVED). Upstream libnl3 will not register conflicting
+		 * capabilities in that range.
+		 *
+		 * Obviously, only backport capability numbers to libnl versions that actually
+		 * implement that capability as well. */
+#undef _NL_SET
+#undef _NL_SETV
+#undef _NL_ASSERT
+	};
+
+	if (capability <= 0 || capability > NL_CAPABILITY_MAX)
+		return 0;
+	capability--;
+	return (caps[capability / 8] & (1 << (capability % 8))) != 0;
+}
+
 /** @endcond */
 
 /** @} */
